#include <stdio.h>
#include "kernels.h"


int main(void)
{
  // initialize csr matrix on host
  int num_rows    = 4;
  int num_cols    = 3;
  int num_entries = 8;
  int   host_row_offsets[]    = {0, 2, 5, 6, 8};
  int   host_column_indices[] = {0, 2, 0, 1, 2, 1, 1, 2};
  float host_values[]         = {1, 2, 3, 4, 5, 6, 7, 8};

  // CSR representation of the following matrix
  // [1 0 2]
  // [3 4 5]
  // [0 6 0]
  // [0 7 8]

  // initialize x and y vectors on host
  float host_x[] = {1, 2, 3};
  float host_y[] = {0, 0, 0, 0};

  // copy csr matrix from host to device
  int   * device_row_offsets;
  int   * device_column_indices;
  float * device_values;
  float * device_x;
  float * device_y;
  cudaMalloc(&device_row_offsets,    (num_rows + 1) * sizeof(int));
  cudaMalloc(&device_column_indices, num_entries    * sizeof(int));
  cudaMalloc(&device_values,         num_entries    * sizeof(float));
  cudaMalloc(&device_x,              num_cols       * sizeof(float));
  cudaMalloc(&device_y,              num_rows       * sizeof(float));
  cudaMemcpy(device_row_offsets,    host_row_offsets,    (num_rows + 1) * sizeof(int),   cudaMemcpyHostToDevice);
  cudaMemcpy(device_column_indices, host_column_indices, num_entries    * sizeof(int),   cudaMemcpyHostToDevice);
  cudaMemcpy(device_values,         host_values,         num_entries    * sizeof(float), cudaMemcpyHostToDevice);
  cudaMemcpy(device_x,              host_x,              num_cols       * sizeof(float), cudaMemcpyHostToDevice);
  cudaMemcpy(device_y,              host_y,              num_rows       * sizeof(float), cudaMemcpyHostToDevice);

  // compute spmv on host
  csr_spmv_host(num_rows, num_cols, num_entries,
                host_row_offsets, host_column_indices, host_values,
                host_x, host_y);
  
  // compute spmv on device
  csr_spmv_device(num_rows, num_cols, num_entries,
                  device_row_offsets, device_column_indices, device_values,
                  device_x, device_y);

  // copy device_y to host
  float * device_y_on_host = (float *) malloc(num_rows * sizeof(float));
  cudaMemcpy(device_y_on_host, device_y, num_rows * sizeof(float), cudaMemcpyDeviceToHost);

  // compare host and device results
  printf("Comparing CSR SpMV results...\n");
  printf("%12s %12s\n", "CPU result", "GPU result");
  for (int row = 0; row < num_rows; row++)
    printf("%12.4f %12.4f\n", host_y[row], device_y_on_host[row]);

  // free memory
  cudaFree(device_row_offsets);
  cudaFree(device_column_indices);
  cudaFree(device_values);
  cudaFree(device_x);        
  cudaFree(device_y);
  free(device_y_on_host);
}
